"use strict";
const tslib_1 = require("tslib");
const command_1 = require("@oclif/command");
const chalk_1 = tslib_1.__importDefault(require("chalk"));
const cli_table_1 = tslib_1.__importDefault(require("cli-table"));
const depcheck_1 = tslib_1.__importDefault(require("depcheck"));
const inquirer_1 = tslib_1.__importDefault(require("inquirer"));
const json_keys_sort_1 = require("json-keys-sort");
const node_emoji_1 = tslib_1.__importDefault(require("node-emoji"));
const path_1 = tslib_1.__importDefault(require("path"));
const util_1 = require("util");
const shelljs_1 = tslib_1.__importDefault(require("shelljs"));
const fs_1 = tslib_1.__importDefault(require("fs"));
const license_checker_1 = tslib_1.__importDefault(require("license-checker"));
const checker = util_1.promisify(license_checker_1.default.init);
const rootPath = path_1.default.resolve();
class Patroller extends command_1.Command {
    async getUnusedDependencies() {
        return new Promise((resolve, reject) => {
            const options = {};
            try {
                depcheck_1.default(rootPath, options, (unused) => {
                    resolve(Object.assign({}, unused));
                });
            }
            catch (error) {
                reject(error);
            }
        });
    }
    async checkDependencies() {
        let uninstallDeps = [];
        let uninstallDevDeps = [];
        const { flags } = this.parse(Patroller);
        const isYarn = flags.yarn || false;
        const unused = await this.getUnusedDependencies();
        const { dependencies, devDependencies } = unused;
        const issues = [...dependencies, ...devDependencies];
        if (issues.length === 0) {
            this.log(chalk_1.default.green(`${node_emoji_1.default.get(':thumbsup:')} no unused dependencies`));
            return;
        }
        uninstallDeps = dependencies || [];
        uninstallDevDeps = devDependencies || [];
        if (!flags.force) {
            this.log(chalk_1.default.yellow(`${node_emoji_1.default.get(':warning:')}  ${issues.length} unused dependencies`));
            const questions = [];
            if (dependencies.length > 0) {
                questions.push({
                    name: 'uninstallDeps',
                    message: 'select the unused dependencies you want to remove',
                    type: 'checkbox',
                    choices: dependencies,
                });
            }
            if (devDependencies.length > 0) {
                questions.push({
                    name: 'uninstallDevDeps',
                    message: 'select the unused devDependencies you want to remove',
                    type: 'checkbox',
                    choices: devDependencies,
                });
            }
            const responses = await inquirer_1.default.prompt(questions);
            uninstallDeps = responses.uninstallDeps || [];
            uninstallDevDeps = responses.uninstallDevDeps || [];
        }
        if (uninstallDeps.length > 0) {
            shelljs_1.default.exec(isYarn
                ? `yarn remove ${uninstallDeps.join(' ')}`
                : `npm uninstall --save ${uninstallDeps.join(' ')}`, { silent: true });
            this.log(chalk_1.default.green(`${node_emoji_1.default.get(':thumbsup:')} uninstalled unused dependencies`));
        }
        if (uninstallDevDeps.length > 0) {
            shelljs_1.default.exec(isYarn
                ? `yarn remove ${uninstallDevDeps.join(' ')}`
                : `npm uninstall --save-dev ${uninstallDevDeps.join(' ')}`, { silent: true });
            this.log(chalk_1.default.green(`${node_emoji_1.default.get(':thumbsup:')} uninstalled unused devDependencies`));
        }
    }
    async lockDependencies() {
        const { flags } = this.parse(Patroller);
        const isYarn = flags.yarn || false;
        const packageJSON = require(path_1.default.join(rootPath, 'package.json'));
        if (packageJSON.dependencies) {
            for (const pkg in packageJSON.dependencies) {
                if (packageJSON.dependencies[pkg]) {
                    packageJSON.dependencies[pkg] = packageJSON.dependencies[pkg].replace('^', '');
                }
            }
        }
        if (packageJSON.devDependencies) {
            for (const pkg in packageJSON.devDependencies) {
                if (packageJSON.devDependencies[pkg]) {
                    packageJSON.devDependencies[pkg] = packageJSON.devDependencies[pkg].replace('^', '');
                }
            }
        }
        fs_1.default.writeFileSync(path_1.default.join(rootPath, 'package.json'), JSON.stringify(packageJSON, null, 2) + '\n');
        shelljs_1.default.exec(isYarn ? 'yarn install' : 'npm install', { silent: true });
        this.log(chalk_1.default.green(`${node_emoji_1.default.get(':thumbsup:')} locked the dependency versions`));
    }
    async sortPackageJSON() {
        const jsonData = require(path_1.default.join(rootPath, 'package.json'));
        const packageJSON = Object.assign(Object.assign({}, jsonData), { dependencies: jsonData.dependencies
                ? json_keys_sort_1.sort(jsonData.dependencies, true)
                : undefined, devDependencies: jsonData.devDependencies
                ? json_keys_sort_1.sort(jsonData.devDependencies, true)
                : undefined });
        fs_1.default.writeFileSync(path_1.default.join(rootPath, 'package.json'), JSON.stringify(packageJSON, null, 2) + '\n');
        this.log(chalk_1.default.green(`${node_emoji_1.default.get(':thumbsup:')} package.json sorted`));
    }
    async checkLicenses() {
        let packages;
        let rootPackage;
        try {
            packages = await checker({ start: rootPath });
        }
        catch (error) {
            this.error(chalk_1.default.red(`${node_emoji_1.default.get(':disappointed:')} cannot find package information in this folder`));
        }
        if (!packages)
            return;
        const table = new cli_table_1.default({ head: ['Package', 'Licenses'] });
        for (const pkg in packages) {
            if (pkg) {
                const p = packages[pkg];
                if (p.path === rootPath) {
                    rootPackage = p;
                }
            }
        }
        const diff = [];
        Object.keys(packages).forEach((key) => {
            if (packages[key] === rootPackage)
                return;
            const matches = packages[key].licenses === rootPackage.licenses;
            const license = matches ? chalk_1.default.dim : chalk_1.default.yellow;
            table.push([key, license(packages[key].licenses)]);
            if (!matches)
                diff.push(packages[key]);
        });
        if (diff.length > 0) {
            this.log(chalk_1.default.yellow(`${node_emoji_1.default.get(':warning:')}  there are some license conflicts`));
            this.log(table.toString());
            return;
        }
        this.log(chalk_1.default.green(`${node_emoji_1.default.get(':thumbsup:')} no license conflicts`));
    }
    async run() {
        const { flags } = this.parse(Patroller);
        try {
            require(path_1.default.join(rootPath, 'package.json'));
        }
        catch (error) {
            this.error(chalk_1.default.red(`${node_emoji_1.default.get(':disappointed:')} cannot find package.json in this folder`));
        }
        // 1) sort the package.json
        if (flags.sort || flags.all) {
            await this.sortPackageJSON();
        }
        // 2) lock the package versions and remove wildcards
        if (flags.lock || flags.all) {
            await this.lockDependencies();
        }
        // 3) check the dependencies for unused using depcheck
        if (flags.unused || flags.all) {
            await this.checkDependencies();
        }
        if (flags.license || flags.all) {
            await this.checkLicenses();
        }
    }
}
Patroller.description = 'A CLI tool for removing unused dependencies, sorting the dependencies, locking the dependencies, and keeping a watch on the licenses.';
Patroller.flags = {
    all: command_1.flags.boolean({ char: 'a', description: 'performs all the actions' }),
    force: command_1.flags.boolean({
        char: 'f',
        description: 'forcefully removes unused dependencies, use with --unused',
    }),
    help: command_1.flags.help({ char: 'h' }),
    license: command_1.flags.boolean({
        description: 'checks the dependencies for difference in license from the root repository',
    }),
    lock: command_1.flags.boolean({
        char: 'l',
        description: 'locks the dependencies in package.json and removes the wildcard ( ^ )',
    }),
    sort: command_1.flags.boolean({
        char: 's',
        description: 'sorts the dependencies in package.json',
    }),
    unused: command_1.flags.boolean({
        char: 'u',
        description: 'removes the unused dependencies from node_modules and the package.json',
    }),
    version: command_1.flags.version({ char: 'v' }),
    yarn: command_1.flags.boolean({ description: 'uses yarn instead of npm' }),
};
module.exports = Patroller;
